#include "libgraph.h"
#import <Cocoa/Cocoa.h>
#include <caml/alloc.h>

const double_t mathPi = 3.14159265358979323846;

extern NSLock* lock;
extern int lineWidth;
extern int width, height;
extern float red, green, blue;
extern BOOL needsRefresh;
extern unsigned char* data;


value caml_gr_plot(value vx, value vy)
{
    int ax = Int_val(vx);
    int ay = Int_val(vy);
    caml_gr_check_open();
    if (ax < 0 || ax >= width || ay < 0 || ay >= height) return Val_unit;
    int k = (height - ay - 1)*width*4+ax*4;
    [lock lock];
    data[k++] = red * 255;
    data[k++] = green * 255;
    data[k++] = blue * 255;
    data[k] = 0xff;
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value caml_gr_moveto(value vx, value vy)
{
    grx = Int_val(vx);
    gry = Int_val(vy);
    return Val_unit;
}

value caml_gr_current_x(void)
{
    return Val_int(grx);
}

value caml_gr_current_y(void)
{
    return Val_int(gry);
}

value caml_gr_lineto(value vx, value vy)
{
    caml_gr_check_open();
    NSPoint p = NSMakePoint(grx, gry);
    grx = Int_val(vx);
    gry = Int_val(vy);
    NSPoint q = NSMakePoint(grx, gry);
    [lock lock];
    @autoreleasepool {
        [NSBezierPath setDefaultLineWidth:lineWidth];
        [NSBezierPath strokeLineFromPoint:p toPoint:q];
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value caml_gr_draw_rect(value vx, value vy, value vw, value vh)
{
    int x = Int_val(vx - 1);
    int y = Int_val(vy - 1);
    int w = Int_val(vw);
    int h = Int_val(vh);
    
    caml_gr_check_open();
    [lock lock];
    @autoreleasepool {
        [NSBezierPath setDefaultLineWidth:lineWidth];
        [NSBezierPath strokeRect: NSMakeRect(x, y, w, h)];
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value caml_gr_draw_arc_nat(value vx, value vy, value vrx, value vry, value va1, value va2)
{
    
    caml_gr_check_open();
    [lock lock];
    @autoreleasepool {
        [[NSGraphicsContext currentContext] saveGraphicsState];
        NSBezierPath* path = [NSBezierPath bezierPath];
        [NSBezierPath setDefaultLineWidth:lineWidth];
        float cx = Int_val(vx);
        float cy = Int_val(vy);
        float r1 = Int_val(vrx);
        float r2 = Int_val(vry);
        float a1 = Int_val(va1);
        float a2 = Int_val(va2);
        NSPoint c = NSMakePoint(cx,cy);
        [path moveToPoint: c];
        float r = r1 + r2;
        [path lineToPoint: NSMakePoint(cx + r * cos(mathPi*a1/180.), cy + r * sin(mathPi*a1/180.))];
        [path appendBezierPathWithArcWithCenter:c radius:r startAngle:a1 endAngle:a2];
        [path closePath];
        [path addClip];
        [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(cx - r1, cy -r2, 2 * r1, 2 * r2)] stroke];
        [[NSGraphicsContext currentContext] restoreGraphicsState];
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value caml_gr_draw_arc(value *argv, int argc)
{
    return caml_gr_draw_arc_nat(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5]);
}

value caml_gr_set_line_width(value vwidth)
{
    caml_gr_check_open();
    lineWidth = Int_val(vwidth);
    return Val_unit;
}
