#import <Cocoa/Cocoa.h>
#import "libgraph.h"
#import "MC.h"
#include <string.h>
#include <caml/callback.h>
#include <caml/fail.h>
#include <unistd.h>

#undef alloc

/*----------------------------------------------------------------------------------------------------*/

//Variables globales

int tailleImage;
int grx, gry;
int orx, ory;
int width;
int height;
NSBitmapImageRep* bitmap = nil;
unsigned char* data = nil;
float red, green, blue;
NSRect frame;
NSString* title = nil;
NSTimer* timer = nil;
char fontName[255];
float fontSize;
int lineWidth;
BOOL needsRefresh = NO;
BOOL newTitle = NO;
BOOL graphIsOpen = NO;
NSLock* lock = nil;        //synchronisation du thread envoyant périodiquement à MacCaml ce qui a été dessiné par Caml.
id <ServerProtocol> mcProxy = nil;
NSMutableArray* queue = nil; //file des événements clavier et souris en provenance de la fenêtre graphique.
NSString* serverName = nil;
NSXPCConnection* connection = nil;
NSXPCListenerEndpoint* endpoint = nil;

/*----------------------------------------------------------------------------------------------------*/

//Classe pour un objet gérant dans un thread la relation avec la vue vendue par MacCaml,
//La boucle interactive de Caml fonctionnant dans un autre thread.

@interface MCClient:NSObject
{
    int b;
}
- (void) mcThread;
- (void) refresh:(NSTimer*) timer;
@end

@implementation MCClient
- (id) init
{
    if (self = [super init])
    {
        b = 0;
    }
    return self;
}

-(void) refresh:(NSTimer*) tmr
{
    if (b == 1)
    {
        usleep(100000);
        b = 0;
    }
    usleep(1000);
    [lock lock];
    if (graphIsOpen == NO)
    {
        if (data) {free(data); data = nil;}
        bitmap = nil;
        [timer invalidate];
        timer = nil;
        [lock unlock];
        queue = nil;
        return;
    }
    if (newTitle == YES)
    {
        [mcProxy setServerFrame:serverName frame:frame title:title];
        newTitle = NO;
        needsRefresh = YES;
    }
    if (needsRefresh == YES)
    {
        [mcProxy getImage:serverName im:[NSData dataWithBytes: data length: tailleImage]];
        needsRefresh = NO;
    }
    __block NSDictionary* e = nil;
l:  [mcProxy getNextEvent:serverName event:^(NSDictionary* evt){e = evt;}];
    usleep(1000);
    if (e == nil)
    {
        [lock unlock];
        return;
    }
    if ((int)[e[@"mask"]intValue] == WINDOWCLOSED)
    {
        graphIsOpen = NO;
        needsRefresh = NO;
        [mcProxy getImage:serverName im:nil];
        [lock unlock];
        return;
    }
    if (queue != nil)
    {
        [queue addObject:e];
        [lock unlock];
        usleep(20000);
        [lock lock];
        e = nil;
        goto l;
    }
    
    [lock unlock];
}

- (void) mcThread
{
    [lock lock];
    b = 1;
    timer = [NSTimer scheduledTimerWithTimeInterval:0.02 target:self selector:@selector(refresh:) userInfo:nil repeats:YES];
    
    [[NSRunLoop currentRunLoop] addTimer:timer forMode:NSDefaultRunLoopMode];
    [lock unlock];
    
    [[NSRunLoop currentRunLoop]run];
}
@end // MCClient

/*----------------------------------------------------------------------------------------------------*/

// Fonctions graphiques de Caml

MCClient* mcClient = nil; //l'objet chargé de la relation avec MacCaml

static value * graphic_failure_exn = NULL;

void caml_gr_fail(char *fmt, char *arg)
{
    char buffer[1024];
    if (graphic_failure_exn == NULL) {
        graphic_failure_exn = caml_named_value("Graphics.Graphic_failure");
        if (graphic_failure_exn == NULL)
            caml_invalid_argument("Exception Graphics.Graphic_failure not initialized, must link graphics.cma");
    }
    sprintf(buffer, fmt, arg);
    caml_raise_with_string(*graphic_failure_exn, buffer);
}

void caml_gr_check_open(void)
{
    [lock lock];
    BOOL b = graphIsOpen;
    [lock unlock];
    if (b == NO)
    {
        [lock lock];
        if (data) {free(data); data = nil;}
        bitmap = nil;
        [timer invalidate];
        timer = nil;
        title = nil;
        [lock unlock];
        queue = nil;
        caml_gr_fail("graphic screen not opened", "");
    }
}

value caml_gr_open_graph(value s)
{
    if (lock == nil) lock = [[NSLock allocWithZone: nil] init];
    
    [lock lock];
    if (graphIsOpen == YES) { [lock unlock]; return Val_unit; }
    graphIsOpen = YES;
    [lock unlock];
    
    orx = 0;
    ory = 0;
    width = 600;
    height = 450;
    
    sscanf(String_val(s), "%dx%d+%d+%d", &width, &height, &orx, &ory);
    frame = NSMakeRect(orx, ory, width, height);
    
    @autoreleasepool {
        title = @"Caml Graphics";
        serverName = [[NSString alloc] initWithCString:getenv("CAMLSERVERNAME") encoding:NSISOLatin1StringEncoding];
        queue =[[NSMutableArray allocWithZone:nil] initWithCapacity:100];
        tailleImage = width*height*4;
        data = (unsigned char*)malloc(tailleImage);
        bitmap = [[NSBitmapImageRep allocWithZone:nil]
                  initWithBitmapDataPlanes:&data
                  pixelsWide:width
                  pixelsHigh:height
                  bitsPerSample:8
                  samplesPerPixel:4 hasAlpha:YES
                  isPlanar:NO
                  colorSpaceName:NSCalibratedRGBColorSpace
                  bytesPerRow:(4 * width)
                  bitsPerPixel:32];
        NSGraphicsContext* gc = nil;
        gc = [NSGraphicsContext graphicsContextWithBitmapImageRep:bitmap];
        [NSGraphicsContext setCurrentContext:gc];
        [gc setShouldAntialias:NO];
        [[NSColor whiteColor] set];
        NSRectFill(NSMakeRect(0,0,width,height));
        [[NSColor blackColor] set];
        [[NSFont fontWithName:@"Courier" size:12] set];
        strcpy(fontName, "Courier");
        needsRefresh = YES;
        newTitle = NO;
        timer = nil;
        grx = 0;
        gry = 0;
        fontSize = 12;
        lineWidth = 0;
        
        [lock lock];
        if (endpoint == nil) {
            NSXPCConnection* connection1 = [[NSXPCConnection alloc] initWithMachServiceName:(NSString *) @"com.jmm.maccamlxpc" options:0];
            connection1.remoteObjectInterface = [NSXPCInterface interfaceWithProtocol:@protocol(helperprotocol)];
            [connection1 resume];
            id <helperprotocol> mcProxy1 = [connection1 remoteObjectProxy];
            if (mcProxy1 == nil)
            {
                NSLog(@"open_graph: connection failure to maccamlxpc");
                graphIsOpen = NO;
                [connection1 invalidate];
                [lock unlock];
                return Val_unit;
            }
            [mcProxy1 listenerEndpoint:^(NSXPCListenerEndpoint* ep){endpoint = ep;}];
            int k = 0;
            while (k < 100 && endpoint == nil){
                usleep(100);
                k++;
            }
            [connection1 invalidate];
            if (endpoint == nil) {
                NSLog(@"open_graph: couldn't get connection endPoint to MacCaml");
                graphIsOpen = NO;
                [lock unlock];
                return Val_unit;
            }
        }
        if (connection == nil) {
            connection = [[NSXPCConnection alloc]initWithListenerEndpoint:endpoint];
            connection.remoteObjectInterface = [NSXPCInterface interfaceWithProtocol:@protocol(ServerProtocol)];
            [connection resume];
            mcProxy = [connection remoteObjectProxy];
            if (mcProxy == nil)
            {
                NSLog(@"open_graph: connection failure to MacCaml");
                graphIsOpen = NO;
                [lock unlock];
                return Val_unit;
            }
        }
        [mcProxy setServerFrame:serverName frame:frame title:title];
        [lock unlock];
        if (mcClient == nil) mcClient = [[MCClient allocWithZone:nil] init];
        [NSThread detachNewThreadSelector:@selector(mcThread) toTarget:mcClient withObject:nil];
    }
    return Val_unit;
}

value caml_gr_close_graph(void)
{
    [lock lock];
    if (graphIsOpen == NO){[lock unlock]; return Val_unit;}
    graphIsOpen = NO;
    needsRefresh = NO;
    
    if (data)
    {
        [[NSColor whiteColor] set];
        NSRectFill(NSMakeRect(0,0,frame.size.width,frame.size.height));
        [mcProxy getImage:serverName im:[NSData dataWithBytes: data length: tailleImage]];
        if (data) {free(data); data = nil;}
    }
    
    bitmap = nil;
    [mcProxy hideWindow:serverName];
    [timer invalidate];
    timer = nil;
    title = nil;
    queue = nil;
    [lock unlock];
    return Val_unit;
}

value caml_gr_window_id(void)
{
    return Val_unit;
}

value caml_gr_open_subwindow(value vx, value vy, value width, value height)
{
    return Val_unit;
}

value caml_gr_close_subwindow(value wid)
{
    return Val_unit;
}

value caml_gr_set_window_title(value s)
{
    [lock lock];
    NSString* t = [NSString stringWithCString:String_val(s) encoding: NSUTF8StringEncoding];
    title = t;
    newTitle = YES;
    [lock unlock];
    return Val_unit;
}

value caml_gr_resize_window (value vx, value vy)
{
    caml_gr_check_open();
    [lock lock];
    if (graphIsOpen == NO){[lock unlock]; return Val_unit;}
    
    if (data)
    {
        [[NSColor whiteColor] set];
        NSRectFill(NSMakeRect(0,0,frame.size.width,frame.size.height));
        [mcProxy getImage:serverName im:[NSData dataWithBytes: data length: tailleImage]];
        if (data) {free(data); data = nil;}
    }
    bitmap = nil;
    [mcProxy hideWindow:serverName];
    [lock unlock];
    queue = nil;
    [lock lock];
    width = Int_val(vx);
    height = Int_val(vy);
    frame = NSMakeRect(orx, ory, width, height);
    
    title = @"Caml Graphics";
    @autoreleasepool {
        queue =[[NSMutableArray allocWithZone:nil] initWithCapacity:100];
        tailleImage = width*height*4;
        data = (unsigned char*)malloc(tailleImage);
        bitmap = [[NSBitmapImageRep allocWithZone:nil]
                  initWithBitmapDataPlanes:&data
                  pixelsWide:width
                  pixelsHigh:height
                  bitsPerSample:8
                  samplesPerPixel:4 hasAlpha:YES
                  isPlanar:NO
                  colorSpaceName:NSCalibratedRGBColorSpace
                  bytesPerRow:(4 * width)
                  bitsPerPixel:32];
        NSGraphicsContext* gc = nil;
        gc = [NSGraphicsContext graphicsContextWithBitmapImageRep:bitmap];
        [NSGraphicsContext setCurrentContext:gc];
        [gc setShouldAntialias:NO];
        [[NSColor whiteColor] set];
        NSRectFill(NSMakeRect(0,0,width,height));
        [[NSColor blackColor] set];
        [[NSFont fontWithName:@"Courier" size:10] set];
        strcpy(fontName, "Courier");
        needsRefresh = YES;
        newTitle = NO;
        timer = nil;
        grx = 0;
        gry = 0;
        fontSize = 10;
        lineWidth = 0;
        [lock unlock];
        [mcProxy setServerFrame:serverName frame:frame title:title];
    }
    return Val_unit;
}

value caml_gr_clear_graph(void)
{
    caml_gr_check_open();
    [lock lock];
    [[NSColor whiteColor] set];
    NSRectFill(NSMakeRect(0,0,frame.size.width,frame.size.height));
    [[NSColor colorWithCalibratedRed: red green: green blue: blue alpha: 1] set];
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value caml_gr_size_x(void)
{
    caml_gr_check_open();
    int w = frame.size.width;
    return Val_int(w);
}

value caml_gr_size_y(void)
{
    caml_gr_check_open();
    int h = frame.size.height;
    return Val_int(h);
}

value caml_gr_synchronize(void)
{
    caml_gr_check_open();
    return Val_unit ;
}

value caml_gr_display_mode(value flag)
{
    caml_gr_check_open();
    return Val_unit ;
}

value caml_gr_remember_mode(value flag)
{
    caml_gr_check_open();
    return Val_unit ;
}

value caml_gr_sigio_signal(value unit)
{
    return Val_int(EVENT_SIGNAL);
}

value caml_gr_sigio_handler(void)
{
    return Val_unit;
}
